from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime

from dataclass_csv import accept_whitespaces, dateformat
from sqlalchemy import Column, DateTime, ForeignKey, Integer
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

import vigorish.database as db
from vigorish.util.dt_format_strings import DATE_ONLY, DATE_ONLY_2, DATE_ONLY_TABLE_ID


class DateScrapeStatus(db.Base):
    __tablename__ = "scrape_status_date"
    id = Column(Integer, primary_key=True)
    game_date = Column(DateTime)
    scraped_daily_dash_bbref = Column(Integer, default=0)
    scraped_daily_dash_brooks = Column(Integer, default=0)
    game_count_bbref = Column(Integer, default=0)
    game_count_brooks = Column(Integer, default=0)
    season_id = Column(Integer, ForeignKey("season.id"))

    games = relationship("GameScrapeStatus", backref="date")
    pitch_apps = relationship("PitchAppScrapeStatus", backref="date")
    pitch_app_status = relationship(
        "Date_PitchApp_View",
        backref="original",
        uselist=False,
        primaryjoin="DateScrapeStatus.id==Date_PitchApp_View.id",
        foreign_keys="Date_PitchApp_View.id",
    )

    @hybrid_property
    def game_date_str(self):
        return self.game_date.strftime(DATE_ONLY_2)

    @hybrid_property
    def total_games(self):
        return len(self.games) if self.games else 0

    @hybrid_property
    def total_games_combined_success(self):
        return sum(game.combined_data_success for game in self.games)

    @hybrid_property
    def total_games_combined_fail(self):
        return sum(game.combined_data_fail for game in self.games)

    @hybrid_property
    def total_games_combined(self):
        return self.total_games_combined_success + self.total_games_combined_fail

    @hybrid_property
    def total_bbref_boxscores_scraped(self):
        return sum(game.scraped_bbref_boxscore for game in self.games)

    @hybrid_property
    def percent_complete_bbref_boxscores_scraped(self):
        return self.total_bbref_boxscores_scraped / float(len(self.games)) if len(self.games) > 0 else 0.0

    @hybrid_property
    def scraped_all_bbref_boxscores(self):
        if not self.scraped_daily_dash_bbref or not self.scraped_daily_dash_brooks:
            return False
        return all(game.scraped_bbref_boxscore == 1 for game in self.games)

    @hybrid_property
    def total_brooks_pitch_logs_scraped(self):
        return sum(game.scraped_brooks_pitch_logs for game in self.games)

    @hybrid_property
    def percent_complete_brooks_pitch_logs(self):
        return self.total_brooks_pitch_logs_scraped / float(len(self.games)) if len(self.games) > 0 else 0.0

    @hybrid_property
    def scraped_all_brooks_pitch_logs(self):
        if not self.scraped_daily_dash_bbref or not self.scraped_daily_dash_brooks:
            return False
        return all(game.scraped_brooks_pitch_logs == 1 for game in self.games)

    @hybrid_property
    def pitch_app_count_bbref(self):
        return sum(game.pitch_app_count_bbref for game in self.games)

    @hybrid_property
    def pitch_app_count_brooks(self):
        return sum(game.pitch_app_count_brooks for game in self.games)

    @hybrid_property
    def total_pitch_count_bbref(self):
        return sum(game.total_pitch_count_bbref for game in self.games)

    @hybrid_property
    def pitch_app_count_pitchfx(self):
        return (
            self.pitch_app_status.total_pitchfx if self.pitch_app_status and self.pitch_app_status.total_pitchfx else 0
        )

    @hybrid_property
    def total_pitch_apps_scraped_pitchfx(self):
        return (
            self.pitch_app_status.total_pitchfx_scraped
            if self.pitch_app_status and self.pitch_app_status.total_pitchfx_scraped
            else 0
        )

    @hybrid_property
    def total_pitch_apps_no_pitchfx_data(self):
        return (
            self.pitch_app_status.total_no_pitchfx_data
            if self.pitch_app_status and self.pitch_app_status.total_no_pitchfx_data
            else 0
        )

    @hybrid_property
    def total_pitch_apps_with_pitchfx_data(self):
        return self.pitch_app_count_pitchfx - self.total_pitch_apps_no_pitchfx_data

    @hybrid_property
    def total_pitch_apps_combined_data(self):
        return (
            self.pitch_app_status.total_combined_pitchfx_bbref_data
            if self.pitch_app_status and self.pitch_app_status.total_combined_pitchfx_bbref_data
            else 0
        )

    @hybrid_property
    def total_pitch_apps_pitchfx_error(self):
        return (
            self.pitch_app_status.total_pitchfx_error
            if self.pitch_app_status and self.pitch_app_status.total_pitchfx_error
            else 0
        )

    @hybrid_property
    def total_pitch_apps_invalid_pitchfx(self):
        return (
            self.pitch_app_status.total_invalid_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_invalid_pitchfx
            else 0
        )

    @hybrid_property
    def total_pitch_apps_pitchfx_is_valid(self):
        return sum(not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.pitch_apps)

    @hybrid_property
    def total_pitch_count_pitch_logs(self):
        return (
            self.pitch_app_status.total_pitch_count_pitch_log
            if self.pitch_app_status and self.pitch_app_status.total_pitch_count_pitch_log
            else 0
        )

    @hybrid_property
    def total_pitch_count_bbref_audited(self):
        return (
            self.pitch_app_status.total_pitch_count_bbref
            if self.pitch_app_status and self.pitch_app_status.total_pitch_count_bbref
            else 0
        )

    @hybrid_property
    def total_pitch_count_pitchfx(self):
        return (
            self.pitch_app_status.total_pitch_count_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_pitch_count_pitchfx
            else 0
        )

    @hybrid_property
    def total_pitch_count_pitchfx_audited(self):
        return (
            self.pitch_app_status.total_pitch_count_pitchfx_audited
            if self.pitch_app_status and self.pitch_app_status.total_pitch_count_pitchfx_audited
            else 0
        )

    @hybrid_property
    def total_missing_pitchfx_count(self):
        return (
            self.pitch_app_status.total_missing_pitchfx_count
            if self.pitch_app_status and self.pitch_app_status.total_missing_pitchfx_count
            else 0
        )

    @hybrid_property
    def total_removed_pitchfx_count(self):
        return (
            self.pitch_app_status.total_removed_pitchfx_count
            if self.pitch_app_status and self.pitch_app_status.total_removed_pitchfx_count
            else 0
        )

    @hybrid_property
    def total_batters_faced_bbref(self):
        return (
            self.pitch_app_status.total_batters_faced_bbref
            if self.pitch_app_status and self.pitch_app_status.total_batters_faced_bbref
            else 0
        )

    @hybrid_property
    def total_batters_faced_pitchfx(self):
        return (
            self.pitch_app_status.total_batters_faced_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_batters_faced_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_pitchfx_complete(self):
        return (
            self.pitch_app_status.total_at_bats_pitchfx_complete
            if self.pitch_app_status and self.pitch_app_status.total_at_bats_pitchfx_complete
            else 0
        )

    @hybrid_property
    def total_at_bats_missing_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_missing_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_at_bats_missing_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_removed_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_removed_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_at_bats_removed_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_pitchfx_error(self):
        return (
            self.pitch_app_status.total_at_bats_pitchfx_error
            if self.pitch_app_status and self.pitch_app_status.total_at_bats_pitchfx_error
            else 0
        )

    @hybrid_property
    def total_at_bats_invalid_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_invalid_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_at_bats_invalid_pitchfx
            else 0
        )

    @hybrid_property
    def scraped_all_pitchfx_logs(self):
        return (
            True
            if not self.pitch_apps
            else False
            if not self.scraped_all_brooks_pitch_logs
            else self.pitch_app_count_pitchfx == self.total_pitch_apps_scraped_pitchfx
        )

    @hybrid_property
    def combined_data_for_all_pitchfx_logs(self):
        return (
            False
            if not self.pitch_apps or not self.scraped_all_pitchfx_logs
            else self.pitch_app_count_pitchfx == self.total_pitch_apps_combined_data
        )

    @hybrid_property
    def pitchfx_error_for_any_pitchfx_logs(self):
        return (
            False
            if not self.pitch_apps or not self.scraped_all_pitchfx_logs
            else any((pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.pitch_apps)
        )

    @hybrid_property
    def pitchfx_is_valid_for_all_pitchfx_logs(self):
        return (
            True
            if not self.pitch_apps or not self.scraped_all_pitchfx_logs
            else all(not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.pitch_apps)
        )

    @hybrid_property
    def percent_complete_pitchfx_logs_scraped(self):
        return (
            self.total_pitch_apps_scraped_pitchfx / float(self.pitch_app_count_pitchfx)
            if self.pitch_app_count_pitchfx > 0
            else 0.0
        )

    @hybrid_property
    def scraped_no_data(self):
        return self.scraped_daily_dash_bbref == 0 and self.scraped_daily_dash_brooks == 0

    @hybrid_property
    def scraped_only_bbref_daily_dash(self):
        return (
            self.scraped_daily_dash_bbref == 1
            and self.scraped_daily_dash_brooks == 0
            and not self.scraped_all_bbref_boxscores
            and not self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scraped_only_brooks_daily_dash(self):
        return (
            self.scraped_daily_dash_bbref == 0
            and self.scraped_daily_dash_brooks == 1
            and not self.scraped_all_bbref_boxscores
            and not self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scraped_only_both_daily_dash(self):
        return (
            self.scraped_daily_dash_bbref == 1
            and self.scraped_daily_dash_brooks == 1
            and not self.scraped_all_bbref_boxscores
            and not self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scraped_only_bbref_boxscores(self):
        return bool(
            self.scraped_all_bbref_boxscores
            and not self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scraped_only_brooks_pitch_logs(self):
        return bool(
            self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_bbref_boxscores
            and not self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scraped_only_both_bbref_boxscores_and_brooks_pitch_logs(self):
        return bool(
            self.scraped_all_bbref_boxscores
            and self.scraped_all_brooks_pitch_logs
            and not self.scraped_all_pitchfx_logs
        )

    @hybrid_property
    def scraped_all_game_data(self):
        return self.scraped_all_bbref_boxscores and self.scraped_all_brooks_pitch_logs and self.scraped_all_pitchfx_logs

    @hybrid_property
    def scrape_status_description(self):
        return (
            "Scraped all game data"
            if self.scraped_all_game_data
            else (
                "Missing Brooks pitchfx logs "
                f"({self.total_pitch_apps_scraped_pitchfx}/{self.pitch_app_count_brooks}, "
                f"{self.percent_complete_pitchfx_logs_scraped:.0%})"
            )
            if self.scraped_only_both_bbref_boxscores_and_brooks_pitch_logs
            else "Scraped Brooks pitch logs, missing BBref boxscores"
            if self.scraped_only_brooks_pitch_logs
            else "Scraped BBref boxscores, missing Brooks pitch logs"
            if self.scraped_only_bbref_boxscores
            else "Missing BBref boxscores and Brooks pitch logs"
            if self.scraped_only_both_daily_dash
            else "Scraped Brooks daily dashboard, missing BBref daily dash"
            if self.scraped_only_brooks_daily_dash
            else "Scraped BBref daily dashboard, missing Brooks daily dash"
            if self.scraped_only_bbref_daily_dash
            else "No data has been scraped"
            if self.scraped_no_data
            else "N/A"
        )

    def __init__(self, **kwargs):
        super(DateScrapeStatus, self).__init__(**kwargs)
        self.id = self.game_date.strftime(DATE_ONLY_TABLE_ID)

    def __repr__(self):
        return f"<DateScrapeStatus date={self.game_date_str}, season_id={self.season_id}>"

    def as_dict(self):
        columns_dict = {c.name: getattr(self, c.name) for c in self.__table__.columns}
        hybrid_properties_dict = {
            "game_date_str": self.game_date_str,
            "total_games": self.total_games,
            "total_games_combined_success": self.total_games_combined_success,
            "total_games_combined_fail": self.total_games_combined_fail,
            "total_games_combined": self.total_games_combined,
            "total_bbref_boxscores_scraped": self.total_bbref_boxscores_scraped,
            "percent_complete_bbref_boxscores_scraped": (f"{self.percent_complete_bbref_boxscores_scraped:01.0f}%"),
            "scraped_all_bbref_boxscores": self.scraped_all_bbref_boxscores,
            "total_brooks_pitch_logs_scraped": self.total_brooks_pitch_logs_scraped,
            "percent_complete_brooks_pitch_logs": f"{self.percent_complete_brooks_pitch_logs:01.0f}%",
            "scraped_all_brooks_pitch_logs": self.scraped_all_brooks_pitch_logs,
            "pitch_app_count_bbref": self.pitch_app_count_bbref,
            "pitch_app_count_brooks": self.pitch_app_count_brooks,
            "total_pitch_count_bbref": self.total_pitch_count_bbref,
            "pitch_app_count_pitchfx": self.pitch_app_count_pitchfx,
            "total_pitch_apps_scraped_pitchfx": self.total_pitch_apps_scraped_pitchfx,
            "total_pitch_apps_no_pitchfx_data": self.total_pitch_apps_no_pitchfx_data,
            "total_pitch_apps_with_pitchfx_data": self.total_pitch_apps_with_pitchfx_data,
            "total_pitch_apps_combined_data": self.total_pitch_apps_combined_data,
            "total_pitch_apps_pitchfx_error": self.total_pitch_apps_pitchfx_error,
            "total_pitch_apps_invalid_pitchfx": self.total_pitch_apps_invalid_pitchfx,
            "total_pitch_apps_pitchfx_is_valid": self.total_pitch_apps_pitchfx_is_valid,
            "total_pitch_count_pitch_logs": self.total_pitch_count_pitch_logs,
            "total_pitch_count_bbref_audited": self.total_pitch_count_bbref_audited,
            "total_pitch_count_pitchfx": self.total_pitch_count_pitchfx,
            "total_pitch_count_pitchfx_audited": self.total_pitch_count_pitchfx_audited,
            "total_missing_pitchfx_count": self.total_missing_pitchfx_count,
            "total_removed_pitchfx_count": self.total_removed_pitchfx_count,
            "total_batters_faced_bbref": self.total_batters_faced_bbref,
            "total_batters_faced_pitchfx": self.total_batters_faced_pitchfx,
            "total_at_bats_pitchfx_complete": self.total_at_bats_pitchfx_complete,
            "total_at_bats_missing_pitchfx": self.total_at_bats_missing_pitchfx,
            "total_at_bats_removed_pitchfx": self.total_at_bats_removed_pitchfx,
            "total_at_bats_pitchfx_error": self.total_at_bats_pitchfx_error,
            "total_at_bats_invalid_pitchfx": self.total_at_bats_invalid_pitchfx,
            "scraped_all_pitchfx_logs": self.scraped_all_pitchfx_logs,
            "combined_data_for_all_pitchfx_logs": self.combined_data_for_all_pitchfx_logs,
            "pitchfx_error_for_any_pitchfx_logs": self.pitchfx_error_for_any_pitchfx_logs,
            "pitchfx_is_valid_for_all_pitchfx_logs": self.pitchfx_is_valid_for_all_pitchfx_logs,
            "percent_complete_pitchfx_logs_scraped": self.percent_complete_pitchfx_logs_scraped,
            "scraped_no_data": self.scraped_only_bbref_daily_dash,
            "scraped_only_bbref_daily_dash": self.scraped_only_bbref_daily_dash,
            "scraped_only_brooks_daily_dash": self.scraped_only_brooks_daily_dash,
            "scraped_only_both_daily_dash": self.scraped_only_both_daily_dash,
            "scraped_only_bbref_boxscores": self.scraped_only_bbref_boxscores,
            "scraped_only_brooks_pitch_logs": self.scraped_only_brooks_pitch_logs,
            "scraped_only_both_bbref_boxscores_and_brooks_pitch_logs": (
                self.scraped_only_both_bbref_boxscores_and_brooks_pitch_logs
            ),
            "scraped_all_game_data": self.scraped_all_game_data,
            "scrape_status_description": self.scrape_status_description,
        }
        return {**columns_dict, **hybrid_properties_dict}

    def status_report(self):
        scraped_daily_bbref = "YES" if self.scraped_daily_dash_bbref == 1 else "NO"
        scraped_daily_brooks = "YES" if self.scraped_daily_dash_brooks == 1 else "NO"
        scraped_bbref_boxscores = "YES" if self.scraped_all_bbref_boxscores else "NO"
        scraped_brooks_pitch_logs = "YES" if self.scraped_all_brooks_pitch_logs else "NO"
        scraped_all_pitchfx_logs = "YES" if self.scraped_all_pitchfx_logs else "NO"
        combined_data_for_all_pitchfx_logs = "YES" if self.combined_data_for_all_pitchfx_logs else "NO"
        pitchfx_error_for_any_pitchfx_logs = "YES" if self.pitchfx_error_for_any_pitchfx_logs else "NO"
        return [
            f"Overall Status For Date......................: {self.scrape_status_description}",
            f"Scraped Daily Dashboard (BBRef/Brooks).......: {scraped_daily_bbref}/{scraped_daily_brooks}",
            (
                f"BBref Boxscores Scraped......................: {scraped_bbref_boxscores} "
                f"{self.total_bbref_boxscores_scraped}/{self.total_games}"
            ),
            (
                f"Brooks Games Scraped.........................: {scraped_brooks_pitch_logs} "
                f"{self.total_brooks_pitch_logs_scraped}/{self.total_games}"
            ),
            (
                f"PitchFx Logs Scraped.........................: {scraped_all_pitchfx_logs} "
                f"{self.total_pitch_apps_scraped_pitchfx}/{self.pitch_app_count_pitchfx} "
                f"({self.percent_complete_pitchfx_logs_scraped:.0%})"
            ),
            (
                f"Combined BBRef/PitchFX Data (Success/Total)..: "
                f"{combined_data_for_all_pitchfx_logs} "
                f"{self.total_games_combined_success}/{self.total_games_combined}"
            ),
            (
                f"Pitch App Count (BBRef/Brooks)...............: "
                f"{self.pitch_app_count_bbref}/{self.pitch_app_count_brooks}"
            ),
            (
                f"Pitch App Count (PFx/data/no data)...........: {self.pitch_app_count_pitchfx}/"
                f"{self.total_pitch_apps_with_pitchfx_data}/{self.total_pitch_apps_no_pitchfx_data}"
            ),
            (
                f"PitchFX Data Errors (Valid AB/Invalid AB)....: "
                f"{pitchfx_error_for_any_pitchfx_logs} "
                f"{self.total_pitch_apps_pitchfx_error}/{self.total_pitch_apps_invalid_pitchfx}"
            ),
            (
                f"Pitch Count (BBRef/Brooks/PFx)...............: {self.total_pitch_count_bbref}/"
                f"{self.total_pitch_count_pitch_logs}/{self.total_pitch_count_pitchfx}"
            ),
            (
                "Pitch Count Audited (BBRef/PFx/Removed)......: "
                f"{self.total_pitch_count_bbref_audited}/{self.total_pitch_count_pitchfx_audited}/"
                f"{self.total_removed_pitchfx_count}"
            ),
        ]

    def games_status_report(self):
        return {g.bbref_game_id: g.status_report() for g in self.games}

    @classmethod
    def find_by_date(cls, db_session, game_date):
        date_str = game_date.strftime(DATE_ONLY_TABLE_ID)
        return db_session.query(cls).get(int(date_str))

    @classmethod
    def get_all_bbref_game_ids_for_date(cls, db_session, game_date):
        return [game.bbref_game_id for game in cls.find_by_date(db_session, game_date).games]

    @classmethod
    def get_all_bbref_scraped_dates_for_season(cls, db_session, season_id):
        return [
            date_status.game_date
            for date_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if date_status.scraped_daily_dash_bbref == 1
        ]

    @classmethod
    def get_all_brooks_scraped_dates_for_season(cls, db_session, season_id):
        return [
            date_status.game_date
            for date_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if date_status.scraped_daily_dash_brooks == 1
        ]

    @classmethod
    def get_unscraped_pitch_app_ids_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        unscraped_pitch_apps = [pitch_app for pitch_app in date_status.pitch_apps if pitch_app.scraped_pitchfx == 0]
        pitch_app_dict = defaultdict(list)
        for pa in unscraped_pitch_apps:
            pitch_app_dict[pa.bbref_game_id].append(pa.pitch_app_id)
        return pitch_app_dict

    @classmethod
    def verify_bbref_daily_dashboard_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_daily_dash_bbref == 1 if date_status else False

    @classmethod
    def verify_brooks_daily_dashboard_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_daily_dash_brooks == 1 if date_status else False

    @classmethod
    def verify_all_bbref_boxscores_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_all_bbref_boxscores if date_status else False

    @classmethod
    def verify_all_brooks_pitch_logs_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_all_brooks_pitch_logs if date_status else False

    @classmethod
    def verify_all_brooks_pitchfx_scraped_for_date(cls, db_session, game_date):
        date_status = cls.find_by_date(db_session, game_date)
        return date_status.scraped_all_pitchfx_logs if date_status else False


@accept_whitespaces
@dateformat(DATE_ONLY)
@dataclass
class DateScrapeStatusCsvRow:
    id: int
    game_date: datetime = None
    scraped_daily_dash_bbref: int = 0
    scraped_daily_dash_brooks: int = 0
    game_count_bbref: int = 0
    game_count_brooks: int = 0



"""Menu item that returns the user to the previous menu."""
import subprocess

from halo import Halo

from vigorish.cli.components import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_heading,
    print_message,
)
from vigorish.enums import DataSet, SyncDirection, VigFile
from vigorish.tasks.base import Task
from vigorish.tasks.sync_scraped_data import SyncScrapedDataTask
from vigorish.util.result import Result

SYNC_STATUS_TEXT_COLOR = {
    "out_of_sync": "bright_green",
    "in_sync": "blue",
    "sync_complete": "bright_green",
    "error": "bright_red",
}


class SyncDataNoPromptsTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.s3_sync = SyncScrapedDataTask(self.app)
        self.sync_direction = None
        self.year = None
        self.file_type = None
        self.data_sets_int = 0
        self.data_set = None
        self.task_number = 0
        self.sync_files = {}
        self.sync_results = []
        self.spinners = {}
        self.results = {}

    @property
    def data_sets(self):
        if self.file_type == VigFile.COMBINED_GAME_DATA:
            return [DataSet.ALL]
        return sorted(ds for ds in DataSet if ds in self.valid_data_sets() and self.data_sets_int & ds == ds)

    @property
    def total_tasks(self):
        return len(self.data_sets)

    @property
    def all_files_are_in_sync(self):
        return all(not out_of_sync for (out_of_sync, _, _) in self.sync_files.values())

    def execute(self, sync_direction, year, file_type, data_sets_int):
        self.sync_direction = sync_direction
        self.year = year
        self.file_type = file_type
        self.data_sets_int = data_sets_int
        self.subscribe_to_events()
        for data_set in self.data_sets:
            self.task_number += 1
            self.data_set = data_set
            self.report_sync_results()
            self.results[data_set] = self.s3_sync.execute(sync_direction, file_type, data_set, year)
            if self.results[data_set].failure:
                return self.results
            self.spinners[data_set].stop()
        self.report_sync_results()
        self.unsubscribe_from_events()
        return self.results

    def valid_data_sets(self):
        data_set_file_type_map = {
            VigFile.SCRAPED_HTML: list(DataSet),
            VigFile.PARSED_JSON: list(DataSet),
            VigFile.COMBINED_GAME_DATA: [DataSet.ALL],
            VigFile.PATCH_LIST: [
                DataSet.BBREF_GAMES_FOR_DATE,
                DataSet.BBREF_BOXSCORES,
                DataSet.BROOKS_GAMES_FOR_DATE,
                DataSet.BROOKS_PITCHFX,
            ],
        }
        return data_set_file_type_map[self.file_type]

    def report_sync_results(self):
        subprocess.run(["clear"])
        self.print_header_message()
        if not self.sync_results:
            return
        for task_result in self.sync_results:
            print_message(task_result[0], wrap=False, fg=task_result[1])

    def print_header_message(self):
        src_folder = "S3 bucket" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "local folder"
        dest_folder = "local folder" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "S3 bucket"
        heading = f"Syncing data from {src_folder} to {dest_folder}"
        print_heading(heading, fg="bright_yellow")

    def error_occurred(self, error_message):
        self.sync_results.append((error_message, SYNC_STATUS_TEXT_COLOR["error"]))
        self.results[self.data_set] = Result.Fail(error_message)

    def get_s3_objects_start(self):
        subprocess.run(["clear"])
        self.spinners["default"] = Halo(
            spinner=get_random_dots_spinner(),
            color=get_random_cli_color(),
            text="Retrieving data for all objects stored in S3...",
        )
        self.spinners["default"].start()

    def get_s3_objects_complete(self):
        self.spinners["default"].stop()

    def find_out_of_sync_files_start(self):
        self.spinners[self.data_set] = Halo(
            spinner=get_random_dots_spinner(),
            color=get_random_cli_color(),
            text=(
                f"Analyzing MLB {self.year} {self.file_type} {self.data_set} files "
                f"(Task {self.task_number}/{self.total_tasks})..."
            ),
        )
        self.spinners[self.data_set].start()

    def find_out_of_sync_files_complete(self, sync_results):
        (out_of_sync, new_files, update_files) = sync_results
        self.sync_files[self.data_set] = (out_of_sync, new_files, update_files)
        sync_files = self.sync_files_text(out_of_sync, new_files, update_files)
        text_color = SYNC_STATUS_TEXT_COLOR["out_of_sync"] if out_of_sync else SYNC_STATUS_TEXT_COLOR["in_sync"]
        self.sync_results.append((sync_files, text_color))

    def sync_files_text(self, out_of_sync, new_files, update_files):
        sync_count = len(new_files) + len(update_files)
        sync_files = f"{sync_count} files out of sync" if out_of_sync else "All files in sync"
        return (
            f"[{self.year} {self.file_type} {self.data_set}] {sync_files} (Task {self.task_number}/{self.total_tasks})"
        )

    def sync_files_start(self, name, complete, total):
        self.report_sync_results()
        self.sync_files_progress(name, complete, total)

    def sync_files_progress(self, name, complete, total):
        direction = "Down" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "Up"
        percent = complete / float(total)
        progress_message = f"{direction}loading: {name} | {percent:.0%} ({complete}/{total} Files)"
        self.spinners[self.data_set].text = progress_message

    def sync_files_complete(self):
        dest_folder = "local folder" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "s3 bucket"
        src_folder = "s3 bucket" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "local folder"
        sync_complete = f"[{self.year} {self.file_type} {self.data_set}] {dest_folder} is in sync with {src_folder}!"
        self.sync_results.append((sync_complete, SYNC_STATUS_TEXT_COLOR["sync_complete"]))

    def subscribe_to_events(self):
        self.s3_sync.events.error_occurred += self.error_occurred
        self.s3_sync.events.get_s3_objects_start += self.get_s3_objects_start
        self.s3_sync.events.get_s3_objects_complete += self.get_s3_objects_complete
        self.s3_sync.events.find_out_of_sync_files_start += self.find_out_of_sync_files_start
        self.s3_sync.events.find_out_of_sync_files_complete += self.find_out_of_sync_files_complete
        self.s3_sync.events.sync_files_start += self.sync_files_start
        self.s3_sync.events.sync_files_progress += self.sync_files_progress
        self.s3_sync.events.sync_files_complete += self.sync_files_complete

    def unsubscribe_from_events(self):
        self.s3_sync.events.error_occurred -= self.error_occurred
        self.s3_sync.events.get_s3_objects_start -= self.get_s3_objects_start
        self.s3_sync.events.get_s3_objects_complete -= self.get_s3_objects_complete
        self.s3_sync.events.find_out_of_sync_files_start -= self.find_out_of_sync_files_start
        self.s3_sync.events.find_out_of_sync_files_complete -= self.find_out_of_sync_files_complete
        self.s3_sync.events.sync_files_start -= self.sync_files_start
        self.s3_sync.events.sync_files_progress -= self.sync_files_progress
        self.s3_sync.events.sync_files_complete -= self.sync_files_complete



from tabulate import tabulate

from vigorish.cli.components.viewers.display_table import DisplayTable
from vigorish.cli.components.viewers.table_viewer import TableViewer


def create_display_table(
    table_rows,
    heading=None,
    message=None,
    table_headers=None,
    tablefmt="fancy_grid",
    numalign="default",
    stralign="default",
    showindex="default",
):
    return DisplayTable(
        table=tabulate(
            table_rows,
            headers=table_headers or (),
            tablefmt=tablefmt,
            numalign=numalign,
            stralign=stralign,
            showindex=showindex,
        ),
        heading=heading,
        message=message,
    )


def create_table_viewer(table_list, table_color="bright_cyan"):
    return TableViewer(
        table_list=table_list,
        prompt="Press Enter to return to previous menu",
        confirm_only=True,
        table_color=table_color,
        heading_color="bright_yellow",
        message_color=None,
    )



"""Menu item that provides detailed information for a single job."""
import subprocess

from getch import pause

from vigorish.cli.components import print_message, user_options_prompt
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.enums import JobStatus
from vigorish.scrape.job_runner import JobRunner
from vigorish.util.list_helpers import report_dict
from vigorish.util.result import Result


class JobDetails(MenuItem):
    def __init__(self, app, db_job, menu_item_number):
        super().__init__(app)
        self.db_job = db_job
        self.job_status = db_job.status
        self.job_details = db_job.job_details
        self.menu_item_text = f" {db_job.name} (Status: {db_job.status.name}, ID: {db_job.id})"
        self.menu_item_emoji = MENU_NUMBERS.get(menu_item_number, f"{menu_item_number}. ")
        self.exit_menu = False

    def launch(self):
        subprocess.run(["clear"])
        print_message("*** Job Details ***", fg="bright_yellow", bold=True)
        job_details = report_dict(self.job_details, title="", title_prefix="", title_suffix="")
        print_message(f"{job_details}\n", wrap=False, fg="bright_yellow")
        if self.db_job.errors:
            print_message("*** Errors ***", fg="bright_red", bold=True)
            print_message(self.db_job.error_messages, fg="bright_red")
        if self.job_status == JobStatus.INCOMPLETE:
            result = self.incomplete_job_options_prompt()
            if result.failure:
                return Result.Ok(self.exit_menu)
            user_choice = result.value
            if user_choice == "RUN":
                job_runner = JobRunner(app=self.app, db_job=self.db_job)
                result = job_runner.execute()
                if result.failure:
                    return result
                return Result.Ok(True)
            if user_choice == "CANCEL":
                self.db_job.status = JobStatus.COMPLETE
                self.db_session.commit()
                subprocess.run(["clear"])
                print_message("Job was successfully cancelled.", fg="bright_red", bold=True)
                pause(message="Press any key to continue...")
                return Result.Ok(True)
        if self.job_status == JobStatus.ERROR:
            result = self.failed_job_options_prompt()
            if result.failure:
                return Result.Ok(self.exit_menu)
            user_choice = result.value
            if user_choice == "RETRY":
                job_runner = JobRunner(app=self.app, db_job=self.db_job)
                result = job_runner.execute()
                if result.failure:
                    return result
                return Result.Ok(True)
        else:
            pause(message="\nPress any key to return to the previous menu...")
            return Result.Ok(self.exit_menu)

    def incomplete_job_options_prompt(self):
        prompt = "\nCurrent options:"
        choices = {
            f"{MENU_NUMBERS.get(1)}  Execute Job": "RUN",
            f"{MENU_NUMBERS.get(2)}  Cancel Job": "CANCEL",
            f"{EMOJIS.get('BACK')} Return to Incomplete Jobs": None,
        }
        return user_options_prompt(choices, prompt, clear_screen=False)

    def failed_job_options_prompt(self):
        prompt = "\nCurrent options:"
        choices = {
            f"{MENU_NUMBERS.get(1)}  Retry Job": "RETRY",
            f"{EMOJIS.get('BACK')} Return to Failed Jobs": None,
        }
        return user_options_prompt(choices, prompt, clear_screen=False)



"""Initialize and update the Player ID Map and Player Team Map CSV files."""
from dataclasses import asdict, dataclass, field

import requests

from vigorish.config.project_paths import PLAYER_ID_MAP_CSV, PLAYER_TEAM_MAP_CSV
from vigorish.tasks.base import Task
from vigorish.util.dataclass_helpers import (
    deserialize_dataclass_list_from_csv,
    get_dataclass_list_from_csv,
    serialize_dataclass_list_to_csv,
)
from vigorish.util.dt_format_strings import DATE_ONLY
from vigorish.util.result import Result

ALL_PITCH_STATS_URL = "https://www.baseball-reference.com/data/war_daily_pitch.txt"
ALL_BAT_STATS_URL = "https://www.baseball-reference.com/data/war_daily_bat.txt"
YEAR_PITCH_STATS_URL = "https://www.baseball-reference.com/data/war_daily_pitch_year.txt"
YEAR_BAT_STATS_URL = "https://www.baseball-reference.com/data/war_daily_bat_year.txt"


@dataclass(eq=True, frozen=True)
class BBRefBatStats:
    name_common: str
    age: str
    mlb_ID: str
    player_ID: str
    year_ID: str
    team_ID: str
    stint_ID: str
    lg_ID: str
    PA: str
    G: str
    Inn: str
    runs_bat: str
    runs_br: str
    runs_dp: str
    runs_field: str
    runs_infield: str
    runs_outfield: str
    runs_catcher: str
    runs_good_plays: str
    runs_defense: str
    runs_position: str
    runs_position_p: str
    runs_replacement: str
    runs_above_rep: str
    runs_above_avg: str
    runs_above_avg_off: str
    runs_above_avg_def: str
    WAA: str
    WAA_off: str
    WAA_def: str
    WAR: str
    WAR_def: str
    WAR_off: str
    WAR_rep: str
    salary: str
    pitcher: str
    teamRpG: str
    oppRpG: str
    oppRpPA_rep: str
    oppRpG_rep: str
    pyth_exponent: str
    pyth_exponent_rep: str
    waa_win_perc: str
    waa_win_perc_off: str
    waa_win_perc_def: str
    waa_win_perc_rep: str
    OPS_plus: str
    TOB_lg: str
    TB_lg: str


@dataclass(eq=True, frozen=True)
class BBRefPitchStats:
    name_common: str
    age: str
    mlb_ID: str
    player_ID: str
    year_ID: str
    team_ID: str
    stint_ID: str
    lg_ID: str
    G: str
    GS: str
    IPouts: str
    IPouts_start: str
    IPouts_relief: str
    RA: str
    xRA: str
    xRA_sprp_adj: str
    xRA_def_pitcher: str
    PPF: str
    PPF_custom: str
    xRA_final: str
    BIP: str
    BIP_perc: str
    RS_def_total: str
    runs_above_avg: str
    runs_above_avg_adj: str
    runs_above_rep: str
    RpO_replacement: str
    GR_leverage_index_avg: str
    WAR: str
    salary: str
    teamRpG: str
    oppRpG: str
    pyth_exponent: str
    waa_win_perc: str
    WAA: str
    WAA_adj: str
    oppRpG_rep: str
    pyth_exponent_rep: str
    waa_win_perc_rep: str
    WAR_rep: str
    ERA_plus: str
    ER_lg: str


@dataclass(eq=True, frozen=True)
class BBRefPlayerIdMap:
    name_common: str = field(compare=False, repr=False)
    mlb_ID: str = field(repr=False)
    player_ID: str


@dataclass(eq=True, frozen=True)
class BBRefPlayerTeamMap:
    name_common: str = field(compare=False, repr=False)
    age: str = field(compare=False, repr=False)
    mlb_ID: str = field(repr=False)
    player_ID: str
    year_ID: str
    team_ID: str
    stint_ID: str = field(repr=False)
    lg_ID: str = field(repr=False)


class UpdatePlayerIdMapTask(Task):
    def __init__(self, app, id_map_filepath=None):
        super().__init__(app)
        if not id_map_filepath:
            id_map_filepath = PLAYER_ID_MAP_CSV
        self.id_map_filepath = id_map_filepath

    def execute(self):
        if not self.id_map_filepath.exists():
            self.initialize_bbref_player_id_map()
            return
        id_map = self.read_bbref_player_id_map_from_file()
        current_id_map = self.get_current_player_id_map()
        new_id_map = list(set(current_id_map) - set(id_map))
        if not new_id_map:
            return Result.Ok([])
        id_map = id_map + new_id_map
        id_map.sort(key=lambda x: x.player_ID)
        self.write_bbref_player_id_map_to_file(id_map)
        new_id_dicts = [asdict(id_map) for id_map in new_id_map] if new_id_map else None
        return Result.Ok(new_id_dicts)

    def initialize_bbref_player_id_map(self):
        if self.id_map_filepath.exists():
            self.id_map_filepath.unlink()
        id_map = self.get_full_player_id_map()
        self.write_bbref_player_id_map_to_file(id_map)

    def get_full_player_id_map(self):
        (bat_stat_list, pitch_stat_list) = parse_full_pitch_and_bat_stats()
        return self.get_player_id_map_from_stats_lists(bat_stat_list, pitch_stat_list)

    def get_current_player_id_map(self):
        (bat_stat_list, pitch_stat_list) = parse_current_pitch_and_bat_stats()
        return self.get_player_id_map_from_stats_lists(bat_stat_list, pitch_stat_list)

    def get_player_id_map_from_stats_lists(self, *stats_lists):
        id_map = {
            BBRefPlayerIdMap(stats.name_common, stats.mlb_ID, stats.player_ID)
            for stats_list in stats_lists
            for stats in stats_list
            if "NULL" not in stats.mlb_ID
        }
        return sorted(id_map, key=lambda x: x.player_ID)

    def write_bbref_player_id_map_to_file(self, new_id_map):
        csv_text = serialize_dataclass_list_to_csv(new_id_map, date_format=DATE_ONLY)
        self.id_map_filepath.write_text(csv_text)

    def read_bbref_player_id_map_from_file(self):
        id_map = get_dataclass_list_from_csv(self.id_map_filepath, BBRefPlayerIdMap)
        return list(set(id_map))


class UpdatePlayerTeamMapTask(Task):
    def __init__(self, app, team_map_filepath=None):
        super().__init__(app)
        if not team_map_filepath:
            team_map_filepath = PLAYER_TEAM_MAP_CSV
        self.team_map_filepath = team_map_filepath

    def execute(self):
        if not self.team_map_filepath.exists():
            self.initialize_bbref_player_team_map()
            return
        team_map = self.read_bbref_player_team_map_from_file()
        current_team_map = self.get_current_team_id_map()
        new_team_map = list(set(current_team_map) - set(team_map))
        if not new_team_map:
            return Result.Ok([])
        team_map = team_map + new_team_map
        team_map.sort(key=lambda x: (x.player_ID, x.year_ID, x.stint_ID))
        self.write_bbref_player_team_map_to_file(team_map)
        new_team_dicts = [asdict(team_map) for team_map in new_team_map] if new_team_map else None
        return Result.Ok(new_team_dicts)

    def initialize_bbref_player_team_map(self):
        if self.team_map_filepath.exists():
            self.team_map_filepath.unlink()
        team_map = self.get_full_team_id_map()
        self.write_bbref_player_team_map_to_file(team_map)

    def get_full_team_id_map(self):
        (bat_stat_list, pitch_stat_list) = parse_full_pitch_and_bat_stats()
        return self.get_player_team_map_from_stats_lists(bat_stat_list, pitch_stat_list)

    def get_current_team_id_map(self):
        (bat_stat_list, pitch_stat_list) = parse_current_pitch_and_bat_stats()
        return self.get_player_team_map_from_stats_lists(bat_stat_list, pitch_stat_list)

    def get_player_team_map_from_stats_lists(self, *stats_lists):
        team_map = {
            BBRefPlayerTeamMap(
                stats.name_common,
                stats.age,
                stats.mlb_ID,
                stats.player_ID,
                stats.year_ID,
                stats.team_ID,
                stats.stint_ID,
                stats.lg_ID,
            )
            for stats_list in stats_lists
            for stats in stats_list
        }
        return sorted(team_map, key=lambda x: (x.player_ID, x.year_ID, x.stint_ID))

    def write_bbref_player_team_map_to_file(self, new_team_map):
        csv_text = serialize_dataclass_list_to_csv(new_team_map, date_format=DATE_ONLY)
        self.team_map_filepath.write_text(csv_text)

    def read_bbref_player_team_map_from_file(self):
        team_map = get_dataclass_list_from_csv(self.team_map_filepath, BBRefPlayerTeamMap)
        return list(set(team_map))


def parse_full_pitch_and_bat_stats():
    pitch_stat_list = parse_pitch_stats(ALL_PITCH_STATS_URL)
    bat_stat_list = parse_bat_stats(ALL_BAT_STATS_URL)
    return (pitch_stat_list, bat_stat_list)


def parse_current_pitch_and_bat_stats():
    pitch_stat_list = parse_pitch_stats(YEAR_PITCH_STATS_URL)
    bat_stat_list = parse_bat_stats(YEAR_BAT_STATS_URL)
    return (pitch_stat_list, bat_stat_list)


def parse_pitch_stats(url):
    response = requests.get(url)
    page_text = response.text
    pitch_stat_list = deserialize_dataclass_list_from_csv(page_text, BBRefPitchStats)
    return list(set(pitch_stat_list))


def parse_bat_stats(url):
    response = requests.get(url)
    page_text = response.text
    bat_stat_list = deserialize_dataclass_list_from_csv(page_text, BBRefBatStats)
    return list(set(bat_stat_list))
